"""
Models for RabbitMQ objects.
"""
import datetime
import json

from typing import Generic, TypeVar, Optional, List, Dict, Any
from pydantic import BaseModel, ConfigDict, field_validator
from pydantic_core import core_schema

TRabbitObject = TypeVar('TRabbitObject', bound='RabbitObject')


def alias_replace_underscore(field_name):
    """Replace '_' with '-' in field name

    Some field names contain '-' which is interpreted as an operator by python, and can
    therefore not be used in field names defined on our pydantic models. Using this as
    an alias_generator we are still able to parse them.
    """
    return field_name.replace('_', '-')


class NonEmptyStr(str):
    def __new__(cls, obj: object):
        s = str(obj)
        if not isinstance(s, str):
            raise ValueError(f'str expected, not {type(s)}')
        if not s:
            raise ValueError('non-empty string empty')
        return str.__new__(NonEmptyStr, s)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: Any) -> Any:
        try:
            return core_schema.no_info_after_validator_function(cls, handler(str))
        except NameError:
            raise RuntimeError(
                "cannot import core_schema from pydantic_core",
            ) from None

    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema: Any, handler: Any) -> Any:
        json_schema = handler(core_schema)
        json_schema.update(minLength=1)
        return json_schema


class RabbitObject(BaseModel, Generic[TRabbitObject]):
    """
    Abstract RabbitMQ object.
    """

    @classmethod
    def from_dict(cls, data):
        return cls(**data)

    @classmethod
    def from_json(cls, json_data):
        data = json.loads(json_data)
        return cls.from_dict(data)


class User(RabbitObject):
    name: str
    password_hash: str
    hashing_algorithm: str
    tags: str


class Metadata(RabbitObject):
    description: str
    tags: List[NonEmptyStr]


class Vhost(RabbitObject):
    name: NonEmptyStr
    description: str
    tags: List[NonEmptyStr]
    metadata: Metadata
    tracing: bool
    cluster_state: Dict[NonEmptyStr, NonEmptyStr]


class Shovel(RabbitObject):
    value: dict
    vhost: NonEmptyStr
    component: NonEmptyStr
    name: NonEmptyStr


class ShovelStatus(RabbitObject):
    node: NonEmptyStr
    timestamp: datetime.datetime
    name: NonEmptyStr
    vhost: NonEmptyStr
    type: NonEmptyStr
    state: NonEmptyStr
    # Only exists when the shovel is in running state
    src_uri: Optional[NonEmptyStr] = None
    src_protocol: Optional[NonEmptyStr] = None
    dest_protocol: Optional[NonEmptyStr] = None
    dest_uri: Optional[NonEmptyStr] = None
    # Only one of each (src_queue, src_exchange and dest_queue, dest_exchange) will be defined
    src_queue: Optional[NonEmptyStr] = None
    src_exchange: Optional[NonEmptyStr] = None
    dest_queue: Optional[NonEmptyStr] = None
    dest_exchange: Optional[NonEmptyStr] = None

    @field_validator('timestamp', mode='before')
    def parse_timestamp(cls, value):
        """Parse timestamp on format YYYY-MM-DD HH:MM:SS or YYYY-MM-DD H:MM:SS"""
        return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")


class Permission(RabbitObject):
    vhost: NonEmptyStr
    user: NonEmptyStr
    configure: NonEmptyStr
    write: NonEmptyStr
    read: NonEmptyStr


class Exchange(RabbitObject):
    arguments: dict
    auto_delete: bool
    durable: bool
    internal: bool
    name: str
    type: NonEmptyStr
    user_who_performed_action: NonEmptyStr
    vhost: NonEmptyStr


class Queue(RabbitObject):
    arguments: dict
    auto_delete: bool
    consumers: Optional[int] = None
    durable: bool
    messages: Optional[int] = None
    name: str
    node: NonEmptyStr
    state: Optional[NonEmptyStr] = None
    type: NonEmptyStr
    vhost: NonEmptyStr


class ConnectionDetails(RabbitObject):
    name: str
    peer_host: Optional[NonEmptyStr] = None


class Connection(ConnectionDetails):
    auth_mechanism: Optional[NonEmptyStr] = None
    channel_max: Optional[int] = None
    channels: Optional[int] = None
    client_properties: dict
    host: Optional[NonEmptyStr] = None
    node: NonEmptyStr
    peer_cert_issuer: Optional[NonEmptyStr] = None
    peer_cert_subject: Optional[NonEmptyStr] = None
    peer_cert_validity: Optional[NonEmptyStr] = None
    peer_port: Optional[int] = None
    port: Optional[int] = None
    ssl: Optional[bool] = None
    ssl_cipher: Optional[NonEmptyStr] = None
    ssl_hash: Optional[NonEmptyStr] = None
    ssl_key_exchange: Optional[NonEmptyStr] = None
    ssl_protocol: Optional[NonEmptyStr] = None
    state: Optional[NonEmptyStr] = None
    timeout: Optional[int] = None
    type: NonEmptyStr
    user: NonEmptyStr
    user_who_performed_action: NonEmptyStr


class Channel(RabbitObject):
    acks_uncommitted: int
    confirm: bool
    connection_details: ConnectionDetails
    consumer_count: int
    name: str
    node: NonEmptyStr
    state: NonEmptyStr
    user: NonEmptyStr
    user_who_performed_action: NonEmptyStr
    vhost: NonEmptyStr


class Binding(RabbitObject):
    source: str
    vhost: str
    destination: NonEmptyStr
    destination_type: NonEmptyStr
    routing_key: str
    arguments: dict
    properties_key: NonEmptyStr


class Definitions(RabbitObject):
    rabbit_version: NonEmptyStr
    rabbitmq_version: NonEmptyStr
    product_name: NonEmptyStr
    product_version: NonEmptyStr
    users: List[User]
    permissions: List[Permission]
    topic_permissions: list
    parameters: list
    global_parameters: list
    policies: list


class Policy(RabbitObject):
    vhost: NonEmptyStr
    name: NonEmptyStr
    pattern: NonEmptyStr
    apply_to: Optional[NonEmptyStr] = None
    definition: dict
    priority: Optional[int] = None

    model_config = ConfigDict(alias_generator=alias_replace_underscore,
                              populate_by_name=True)
